#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

import os
import sys
import time
import socket
import struct
from collections import namedtuple

"""
ripcap.py: pcap parser except it's fast --
  in pypy2:    940 MiB/s, 650k UDP-pkt/s, 325k TCP-pkt/s  ☜(ﾟ∀ﾟ☜)
  in pypy3:    402 MiB/s, 279k UDP-pkt/s, 315k TCP-pkt/s
  in cpython3: 139 MiB/s,  96k UDP-pkt/s,  62k TCP-pkt/s
  in cpython2: 104 MiB/s,  72k UDP-pkt/s,  45k TCP-pkt/s

changes in v2.1:
 * frame.pcap_pos is absolute offset in pcap file
 * frame.data_ofs is offset from pcap_pos to payload
 * speed indicator includes total speed as well
 * option to ignore unimplemented packet types

"""
__version__ = (2, 1)
__date__ = "2023-12-05"
__author__ = "ed <irc.rizon.net>"
__url__ = "https://ocv.me/dev/ripcap.py"
__credits__ = ["stackoverflow.com"]
__license__ = "MIT"
__copyright__ = 2020


# config
ALLOW_UNKNOWN_PCAPNG_BLOCKS = True
ALLOW_UNKNOWN_PCAPNG_BLOCKS = False

IGNORE_UNKNOWN_PACKET_TYPES = True
IGNORE_UNKNOWN_PACKET_TYPES = False


PY2 = sys.version_info[0] == 2
if PY2:
    from cStringIO import StringIO as BytesIO
else:
    from io import BytesIO as BytesIO


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def b2hex(data):
    if PY2:
        return " ".join(map(lambda b: format(ord(b), "02x"), data))
    else:
        if type(data) is str:
            return " ".join(map(lambda b: format(ord(b), "02x"), data))
        else:
            return " ".join(map(lambda b: format(b, "02x"), data))


class PositionTracker(object):
    def __init__(self, buf):
        self.f = buf
        self.pos = 0

    def read(self, n=-1):
        rv = self.f.read(n)
        self.pos += len(rv)
        return rv

    def tell(self):
        return self.pos


class Parser(object):
    def __init__(self, buf):
        if hasattr(buf, "read"):
            try:
                buf.tell()
                self.f = buf
                # eprint('[*] direct buffer access')
            except:
                self.f = PositionTracker(buf)
                # eprint('[*] indirect access using PositionTracker')
        else:
            # eprint('[*] indirect access using BytesIO')
            self.f = BytesIO(buf)

    def read(self, sz):
        ret = self.f.read(sz)
        if len(ret) == sz:
            return ret

        while len(ret) < sz:
            b = self.f.read(sz - len(ret))
            ret += b
            if not b:
                break

        if len(ret) == sz:
            return ret

        raise Exception("EOF")

    def u64(self):
        buf = self.read(8)
        return struct.unpack(b"<Q", buf)[0]

    def ru64(self):
        buf = self.read(8)
        return struct.unpack(b">Q", buf)[0]

    def s32(self):
        buf = self.read(4)
        return struct.unpack(b"<l", buf)[0]

    def u32(self):
        buf = self.read(4)
        return struct.unpack(b"<L", buf)[0]

    def ru32(self):
        buf = self.read(4)
        return struct.unpack(b">L", buf)[0]

    def u16(self):
        buf = self.read(2)
        return struct.unpack(b"<H", buf)[0]

    def ru16(self):
        buf = self.read(2)
        return struct.unpack(b">H", buf)[0]

    def u8(self):
        buf = self.read(1)
        return struct.unpack(b"B", buf)[0]

    def drop(self, n_bytes):
        self.read(n_bytes)

    def unk(self, expect):
        v = self.read(len(expect))
        if v != expect:
            t = "unexpected unknown:\n{0} got\n{1} expected"
            raise Exception(t.format(b2hex(v), b2hex(expect)))

    def str(self, maxlen):
        size = self.u32()
        if size > maxlen:
            raise Exception("string out of bounds: {0}".format(size))

        return self.read(size).decode("utf-8").rstrip("\0")

    def align(self, to):
        pos = self.f.tell()
        waste = 0
        while divmod(pos, to)[1] != 0:
            waste += 1
            pos += 1

        if waste != 0:
            self.read(waste)


class Ripcap(object):
    VERSION = __version__

    def __init__(self, fobj, status_interval=8192):
        self.fobj = fobj
        self.status_interval = status_interval
        self.num_packets = 0
        self.num_bytes = 0
        self.last_num_packets = 0
        self.last_num_bytes = 0
        self.last_print = 0
        self.first_print = 0

    Frame = namedtuple(
        "Frame",
        "pcap_pos data_ofs data header1 header2 src_mac src_ip src_port dst_mac dst_ip dst_port",
    )

    def _get_error_context(self, p):
        ret = ""
        try:
            for n in range(8):
                ln = b2hex(p.f.read(16))
                if not ln:
                    break

                ret += "\n" + ln
        except:
            pass

        return ret

    def handle_packet(self, pcap, buf):
        if self.status_interval and self.num_packets % self.status_interval == 0:
            now = time.time()
            if self.last_print != 0:
                td = now - self.last_print
                td0 = now - self.first_print
                byte_sec = (self.num_bytes - self.last_num_bytes) / td
                pkt_sec = (self.num_packets - self.last_num_packets) / td
                tbyte_sec = self.num_bytes / td0
                tpkt_sec = self.num_packets / td0
            else:
                self.first_print = now
                byte_sec = tbyte_sec = 0
                pkt_sec = tpkt_sec = 0

            self.last_print = now
            self.last_num_bytes = self.num_bytes
            self.last_num_packets = self.num_packets

            m = "\033[1;30m[*] {:.2f}k pkts, {:.2f} MiB, {:.2f}k pk/s, {:.2f} M/s, T {:.2f} {:.2f}\033[0m".format(
                self.num_packets / 1000.0,
                self.num_bytes / (1024.0 * 1024.0),
                pkt_sec / 1000.0,
                byte_sec / (1024.0 * 1024.0),
                tpkt_sec / 1000.0,
                tbyte_sec / (1024.0 * 1024.0),
            )
            eprint(m)

        pcap_pos = pcap.f.tell() - len(buf)

        p = Parser(buf)
        dst_mac = p.read(6)
        src_mac = p.read(6)

        pkt_type = p.u16()
        if pkt_type not in (0x8, 0xDD86):
            if IGNORE_UNKNOWN_PACKET_TYPES:
                return None

            t = "at pcap offset 0x{:x}, packet offset 0x{:x}, expected IPv4 (0x8) or IPv6 (0x86dd), got 0x{:x}, and IGNORE_UNKNOWN_PACKET_TYPES is False"
            raise RuntimeError(t.format(pcap_pos, p.f.tell(), pkt_type))

        if pkt_type == 0x8:
            blk_pos = p.f.tell()
            try:
                # u8 0x45 (ipv4) + header length
                # u8 differentiated services field + explicit congestion notification
                p.drop(2)
                frame_sz = p.ru16()
                seq_no = p.ru16()
            except Exception as e:
                t = "at pcap offset 0x{:x}, packet offset 0x{:x}, got unexpected data"
                raise RuntimeError(t.format(pcap_pos, p.f.tell()))

            flags, frag_ofs = divmod(p.ru16(), 0x2000)
            fragmented = frag_ofs != 0
            if fragmented:
                t = "[x] at pcap offset 0x{:x}, got fragmented packet, will discard"
                eprint(t.format(pcap_pos))

            p.u8()  # TTL
            protocol = p.u8()
            p.u16()  # checksum
            src_ip = socket.inet_ntoa(p.read(4))
            dst_ip = socket.inet_ntoa(p.read(4))
            h1 = {"type": pkt_type, "flags": flags, "seq_no": seq_no}

        else:
            # 4bit version + traffic class
            # 8bit traffic class
            # 5bit flow label
            p.drop(4)
            frame_sz = p.ru16()
            protocol = p.u8()
            p.drop(1)  # hop limit
            src_ip = socket.inet_ntop(socket.AF_INET6, p.read(16))
            dst_ip = socket.inet_ntop(socket.AF_INET6, p.read(16))
            blk_pos = p.f.tell()
            fragmented = False  # ┐(ﾟ∀ﾟ)┌
            h1 = {"type": pkt_type, "flags": None}

        if protocol not in (0x11, 0x06):
            if IGNORE_UNKNOWN_PACKET_TYPES:
                return None

            eprint("[{}] [{}]".format(src_ip, dst_ip))
            raise RuntimeError(
                "at pcap offset 0x{:x}, packet offset 0x{:x}, expected UDP (0x11) or TCP (0x06), got 0x{:x}, and IGNORE_UNKNOWN_PACKET_TYPES is False{}".format(
                    pcap_pos, p.f.tell(), protocol, self._get_error_context(p)
                )
            )

        ofs_p2 = p.f.tell()

        if protocol == 0x11:
            src_port = p.ru16()
            dst_port = p.ru16()
            pkt_len = p.ru16() - 8
            p.u16()  # checksum
            len_chk = blk_pos + frame_sz - p.f.tell()
            if pkt_len != len_chk:
                t = "at pcap offset 0x{:x}, packet offset 0x{:x}, bad packet length: expected 0x{:x} bytes, got 0x{:x}"
                raise RuntimeError(t.format(pcap_pos, p.f.tell(), len_chk, pkt_len))

            h2 = {"type": protocol, "flags": None}

        else:
            src_port = p.ru16()
            dst_port = p.ru16()
            seq_no = p.ru32()
            ack_no = p.ru32()
            data_offset = int(p.u8() / 0x10) * 4
            flags = p.u8()
            opaque = p.read(data_offset - (4 * 3 + 2))
            pkt_len = blk_pos + frame_sz - p.f.tell()
            h2 = {
                "type": protocol,
                "flags": flags,
                "seq_no": seq_no,
                "ack_no": ack_no,
                "opaque": opaque,
            }

        data_ofs = p.f.tell()
        h1["data"] = buf[:ofs_p2]
        h2["data"] = buf[ofs_p2:data_ofs]

        data = p.read(pkt_len)
        if len(data) != pkt_len:
            t = "at pcap offset 0x{:x}, packet offset 0x{:x}, partial packet: expected 0x{:x} bytes, got 0x{:x}"
            raise RuntimeError(t.format(pcap_pos, p.f.tell(), pkt_len, len(data)))

        self.num_packets += 1
        self.num_bytes += len(data)
        if fragmented:
            return None
        else:
            return Ripcap.Frame(
                pcap_pos,
                data_ofs,
                data,
                h1,
                h2,
                src_mac,
                src_ip,
                src_port,
                dst_mac,
                dst_ip,
                dst_port,
            )

    def parse_pcap(self):
        p = Parser(self.fobj)

        bom = p.u32()
        if bom == 0x0A0D0D0A:
            raise RuntimeError("\n\n    this is a pcapng file, not pcap\n")
        if bom != 0xA1B2C3D4:
            raise RuntimeError("unexpected endianness in pcap: {:08x}".format(bom))

        ver_maj = p.u16()
        ver_min = p.u16()
        if ver_maj != 2:
            # tested with 2.4
            t = "unsupported pcap version: {0}.{1}"
            raise RuntimeError(t.format(ver_maj, ver_min))

        p.unk(b"\x00\x00\x00\x00")  # timezone offset in seconds; always GMT
        p.unk(b"\x00\x00\x00\x00")  # time accuracy; always 0

        # u32 max packet length (usually 65535, dontcare anyways)
        # u32 network type (ethernet = 1)
        p.drop(8)

        while True:
            try:
                ts_hi = p.u32()
            except:
                return

            ts_lo = p.u32()
            cap_len = p.u32()
            pkt_len = p.u32()
            if cap_len != pkt_len:
                t = "not impl: partial packet at 0x{:x}; {} {}"
                raise RuntimeError(t.format(p.f.tell(), cap_len, pkt_len))

            buf = p.read(pkt_len)
            ret = self.handle_packet(p, buf)
            if ret:
                yield [ts_hi + ts_lo / (1000.0 * 1000), ret]

    def parse_pcapng(self):
        p = Parser(self.fobj)
        while True:
            blk_pos = p.f.tell()
            try:
                blk_type = p.u32()
                blk_len = p.u32()
            except:
                break  # eof

            if blk_type == 0x0A0D0D0A:  # Section Header Block
                bom = p.u32()
                if bom != 0x1A2B3C4D:
                    raise RuntimeError("unexpected endianness in pcapng")

                ver_maj = p.u16()
                ver_min = p.u16()
                if ver_maj != 1 or ver_min != 0:
                    t = "unsupported pcapng version: {0}.{1}"
                    raise RuntimeError(t.format(ver_maj, ver_min))

            elif blk_type == 0x1:  # Interface Description Block
                pass

            elif blk_type == 0x6:  # Enhanced Packet block
                p.u32()  # interface id
                ts_hi = p.u32()
                ts_lo = p.u32()
                cap_len = p.u32()
                pkt_len = p.u32()
                if cap_len != pkt_len:
                    t = "not impl: partial packet at 0x{:x}"
                    raise RuntimeError(t.format(p.f.tell()))

                buf = p.read(pkt_len)
                p.align(4)
                ret = self.handle_packet(p, buf)
                if ret:
                    yield [ts_hi + ts_lo / (1000.0 * 1000 * 1000), ret]

            elif blk_type == 0x3:  # Simple Packet block (UNTESTED)
                pkt_len = p.u32()
                buf = p.read(pkt_len)
                p.align(4)
                ret = self.handle_packet(p, buf)
                if ret:
                    yield [time.time(), ret]

            elif blk_type == 0x5:  # Interface Statistics Block
                pass

            elif blk_type == 0xA1B2C3D4:
                raise RuntimeError("\n\n    this is a pcap file, not pcapng\n")

            else:
                if not ALLOW_UNKNOWN_PCAPNG_BLOCKS:
                    t = "unsupported pcapng block type: {:x}"
                    raise RuntimeError(t.format(blk_type))

            # final 4 bytes of each block is another repetition of blk_len,
            # use this as a sanity check
            p.drop((blk_len - (p.f.tell() - blk_pos)) - 4)
            v = p.u32()
            if v != blk_len:
                t = "corrupt pcapng; expected {:x}, got {:x}"
                raise RuntimeError(t.format(blk_len, v))


class Tcpdump(object):
    def __init__(self, iface, cap_filter="", log_file=None):
        import subprocess as sp

        p = sp.Popen(["tcpdump", "--help"], stdout=sp.PIPE, stderr=sp.PIPE)
        chk = p.communicate()
        have_im = b"--immediate-mode" in b"".join(chk)

        cmd = ["tcpdump", "-Unni", str(iface), "-w", "-"]

        if have_im:
            cmd.append("--immediate-mode")

        if cap_filter:
            cmd.append(str(cap_filter))

        eprint("\033[1;30m[*] " + " ".join(cmd) + "\033[0m")
        self.p = sp.Popen(cmd, stdout=sp.PIPE)

        if False:
            self.buf = b""
            self.read = self.buggy_read
        elif not log_file:
            self.read = self.p.stdout.read
        else:
            self.read = self.log_read
            self.log_fh = open(log_file, "wb")

    def log_read(self, sz):
        ret = self.p.stdout.read(sz)
        while not ret:
            if self.p.poll() is not None:
                return b""

            time.sleep(0.01)
            ret = self.p.stdout.read(sz)

        if self.log_fh:
            self.log_fh.write(ret)

        return ret

    def buggy_read(self, sz):
        # eprint("read", sz)
        n = min(sz, 13)
        if len(self.buf) < n:
            self.buf += self.p.stdout.read(n)

        ret = self.buf[:n]
        self.buf = self.buf[n:]
        return ret


def main():
    import re
    import argparse

    ap = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
examples:
  ./ripcap.py -v -i lo -f 'port 4321'
  tcpdump -Unni lo --immediate-mode -w - | ./ripcap.py -v
""",
    )

    ap.add_argument("-ng", action="store_true", help="assume pcapng input")
    ap.add_argument("-v", action="store_true", help="print packet contents")
    ap.add_argument("-i", metavar="IFACE", type=str, help="interface to capture")
    ap.add_argument("-f", metavar="FILTER", type=str, help="tcpdump filter")
    ar = ap.parse_args()

    if not PY2:
        stdin = sys.stdin.buffer
    else:
        if sys.platform == "win32":
            import msvcrt

            msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)

        stdin = sys.stdin

    src = "stdin"
    src_fobj = stdin
    if ar.i:
        src = "tcpdump subprocess"
        src_fobj = Tcpdump(ar.i, ar.f)

    if ar.ng:
        eprint("[*] reading pcapng from", src)
    else:
        eprint("[*] reading pcap from", src, "(use -ng for pcapng input)")

    if not ar.i:
        eprint("[*] see -h for examples and more")
        eprint()

    p = Ripcap(src_fobj)
    pp = p.parse_pcapng if ar.ng else p.parse_pcap

    if PY2:
        ccs = "".join(map(unichr, range(0x00, 0x20) + range(0x7F, 0xA0)))
    else:
        import itertools

        ccs = "".join(map(chr, itertools.chain(range(0x00, 0x20), range(0x7F, 0xA0))))

    cc_re = re.compile("[%s]" % re.escape(ccs))

    for ts, pkt in pp():
        # retrieve some parameters for a more realistic
        # performance estimate, even if non-verbose
        if "%s %s %d" % (pkt.src_ip, pkt.dst_port, len(pkt.data)) and not ar.v:
            continue

        print(
            "[*] %.3f  %d+%d  %15s/%5d  %15s/%5d  %s,%s,%02x  |%5d|  [%s]"
            % (
                ts,
                pkt.pcap_pos,
                pkt.data_ofs,
                pkt.src_ip,
                pkt.src_port,
                pkt.dst_ip,
                pkt.dst_port,
                pkt.header1["type"],
                pkt.header2["type"],
                pkt.header2["flags"] or 0,
                len(pkt.data),
                cc_re.sub("", pkt.data.decode("utf-8", "replace")[:64]),
            )
        )

    eprint("[+] done")


if __name__ == "__main__":
    main()


"""
py3 example (save to same folder as this):

#!/usr/bin/env python3
from ripcap import Ripcap, Tcpdump, sys
from pprint import pprint

# p = Ripcap(sys.stdin.buffer)
p = Ripcap(Tcpdump("lo", "port 4321"))
for ts, frame in p.parse_pcap():
    pprint([ts, dict(frame._asdict())])



# generate packets
g() { while true; do date +%s.%N; sleep 0.1; done | stdbuf -o0 tr -d '\n'; }

# udp server/client:
socat udp4-recv:4321,reuseaddr -
g | socat - udp4-sendto:127.0.0.1:4321

# tcp server/client:
socat tcp-listen:4321,fork,reuseaddr -
g | socat - tcp-connect:127.0.0.1:4321

# tcp6:
socat tcp6-listen:4321,fork,reuseaddr -
g | socat - tcp-connect:'[::1]':4321

# udp6:
socat udp6-listen:4321,fork,reuseaddr -
g | socat - udp6-connect:'[::1]':4321

"""
